#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <regex.h>
#include <pthread.h>

#include <simclist.h>


#include "sshguard_log.h"
#include "sshguard_fw.h"
#include "sshguard.h"

#define MAX_LOGLINE_LEN     500
#define IP_LEN              16

list_t limbo;
list_t hell;

typedef struct {
    char ip[IP_LEN];
    time_t whenfirst;
    time_t whenlast;
    unsigned int numhits;
} ipentry_t;

struct {
    time_t pardon_threshold, stale_threshold;
    unsigned int abuse_threshold;
} opts;

static inline void extractIP(char *origline, char *ipstr, off_t start, off_t end);
static inline void ipentryinit(ipentry_t *ipe, char *ipaddr);
static void usage(void);
void reportIP(char *ipaddr);
void purge_limbo_stale(void);
void *pardonBlocked(void *par);
void sighand(int sig);

regex_t crackre;

int main(int argc, char *argv[]) {
    char ip[IP_LEN], logline[MAX_LOGLINE_LEN];
    regmatch_t pmatch[REGEXIPENTRY+1];
    pthread_t tid;
    int retv;
    

    retv = regcomp(&crackre, SSHCRACK_REGEX, REG_EXTENDED);
    if (retv != 0) {
        regerror(retv, &crackre, logline, MAX_LOGLINE_LEN);
        fprintf(stderr, "%s\n", logline);
        exit(1);
    }

    list_init(&limbo);
    list_init(&hell);

    sshguard_log_init();
    if (fw_init() != 0) {
        fprintf(stderr, "Could not init firewall.\n");
        exit(1);
    }


    /* parsing the command line */
    opts.pardon_threshold = DEFAULT_PARDON_THRESHOLD;
    opts.stale_threshold = DEFAULT_STALE_THRESHOLD;
    opts.abuse_threshold = DEFAULT_ABUSE_THRESHOLD;
    while ((ip[0] = getopt(argc, argv, "p:s:a:")) != -1) {
        switch (ip[0]) {
            case 'p':   /* pardon threshold interval */
                opts.pardon_threshold = strtol(optarg, (char **)NULL, 10);
                break;
            case 's':   /* stale threshold interval */
                opts.stale_threshold = strtol(optarg, (char **)NULL, 10);
                break;
            case 'a':   /* abuse threshold count */
                opts.abuse_threshold = strtol(optarg, (char **)NULL, 10);
                break;
            default:
                usage();
                exit(1);
        }
    }

    /* set signal handler */
    signal(SIGTERM, sighand);
    signal(SIGINT, sighand);
    signal(SIGQUIT, sighand);

    if (pthread_create(&tid, NULL, pardonBlocked, NULL) != 0) {
        perror("pthread_create()");
        exit(2);
    }
    
    sshguard_log(LOG_INFO, "Started successfully [(a,p,s)=(%u, %u, %u)], now ready to scan.", \
            opts.abuse_threshold, (unsigned int)opts.pardon_threshold, (unsigned int)opts.stale_threshold);


    while (fgets(logline, MAX_LOGLINE_LEN, stdin) != NULL) {
        if (regexec(&crackre, logline, REGEXIPENTRY+1, pmatch, 0) != 0) continue;

        /* extract the IP address */
        extractIP(logline, ip, pmatch[REGEXIPENTRY].rm_so, pmatch[REGEXIPENTRY].rm_eo);
        sshguard_log(LOG_DEBUG, "Matched IP address %s\n", ip);
        
        /* report IP */
        reportIP(ip);
    }

    /* cleanup */
    regfree(&crackre);
    if (fw_fin() != 0) sshguard_log(LOG_ERR, "Cound not finalize firewall.");

    sshguard_log_fin();

    return 0;
}

static inline void extractIP(char *origline, char *ipstr, off_t start, off_t end) {
    off_t ind, j;
    
    for (j = 0, ind = start; ind < end; j++, ind++) {
        ipstr[j] = origline[ind];
    }
    ipstr[j] = '\0';
}

void reportIP(char *ipaddr) {
    ipentry_t *tmpent = NULL;
    unsigned int pos;
    int ret;


    /* clean list from stale entries */
    purge_limbo_stale();
    
    /* search entry in list */
    pos = 0;
    list_iterator_start(&limbo);
    while (list_iterator_hasnext(&limbo)) {
        tmpent = list_iterator_next(&limbo);
        if (strcmp(tmpent->ip, ipaddr) == 0) /* found! */
            break;
        pos++;
    }
    list_iterator_stop(&limbo);
    
    if (pos < list_size(&limbo)) { /* then found */
        /* update last hit */
        tmpent->whenlast = time(NULL);
        tmpent->numhits++;
        if (tmpent->numhits >= opts.abuse_threshold) { /* block IP */
            sshguard_log(LOG_NOTICE, "Blocking %s: %u failures over %u seconds.\n", tmpent->ip, tmpent->numhits, tmpent->whenlast - tmpent->whenfirst);
            /* fprintf(stderr, "Blocking %s: %u failures over %u seconds.\n", tmpent->ip, tmpent->numhits, tmpent->whenlast - tmpent->whenfirst); */
            ret = block_ip(ipaddr);
            if (ret != 0) sshguard_log(LOG_ERR, "Blocking command failed. Exited: %d", ret);
            /* re-get elem for deleting it from limbo list */
            tmpent = list_extract_at(& limbo, pos);
            list_append(&hell, tmpent);
        }
        return;
    }
    
    /* otherwise: insert the new item */
    tmpent = malloc(sizeof(ipentry_t));
    ipentryinit(tmpent, ipaddr);
    list_append(&limbo, tmpent);
}

static inline void ipentryinit(ipentry_t *ipe, char *ipaddr) {
    strcpy(ipe->ip, ipaddr);
    ipe->whenfirst = ipe->whenlast = time(NULL);
    ipe->numhits = 1;
}

void purge_limbo_stale(void) {
    ipentry_t *tmpent;
    time_t now;
    unsigned int pos = 0;


    now = time(NULL);
    for (pos = 0; pos < list_size(&limbo); pos++) {
        tmpent = list_get_at(&limbo, pos);
        if (now - tmpent->whenfirst > opts.stale_threshold)
            list_delete_at(&limbo, pos);
    }
}

void *pardonBlocked(void *par) {
    time_t now;
    ipentry_t *tmpel;
    int ret;


    srandom(time(NULL));

    while (1) {
        /* wait some time, at most opts.pardon_threshold/3 + 1 sec */
        sleep(1 + random() % (opts.pardon_threshold/2));
        now = time(NULL);
        tmpel = list_get_at(&hell, 0);
        while (tmpel != NULL && now - tmpel->whenlast > opts.pardon_threshold) {
            sshguard_log(LOG_INFO, "Releasing %s after %u seconds.\n", tmpel->ip, now - tmpel->whenlast);
            /* fprintf(stderr, "Releasing %s after %u seconds.\n", tmpel->ip, now - tmpel->whenlast); */
            ret = release_ip(tmpel->ip);
            if (ret != 0) sshguard_log(LOG_ERR, "Release command failed. Exited: %d", ret);
            list_delete_at(&hell, 0);
            free(tmpel);
            tmpel = list_get_at(&hell, 0);
        }
    }

    pthread_exit(NULL);
    return NULL;
}

void sighand(int sig) {
    /* flush blocking rules */
    sshguard_log(LOG_INFO, "Got exit signal, flushing blocked addresses and exiting...");
    flush_blocked();
    if (fw_fin() != 0) sshguard_log(LOG_ERR, "Cound not finalize firewall.");
    closelog();

    regfree(&crackre);

    exit(0);
}

static void usage(void) {
    fprintf(stderr, "Usage:\nsshguard [-a] [-p] [-s]\n");
    fprintf(stderr, "\t-a\tNumber of hits after which blocking an address (%d)\n", DEFAULT_ABUSE_THRESHOLD);
    fprintf(stderr, "\t-p\tSeconds after which unblocking a blocked address (%d)\n", DEFAULT_PARDON_THRESHOLD);
    fprintf(stderr, "\t-s\tSeconds after which forgetting about a cracker candidate (%d)\n", DEFAULT_STALE_THRESHOLD);
}

